from unlok import Unlok
from unlok.rath import UnlokRath, UnlokLinkComposition
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.contrib.herre.links.auth import HerreAuthLink
from herre import Herre
from fakts import Fakts


class ArkitektUnlok(Unlok):
    rath: UnlokRath


def build_arkitekt_unlok(herre: Herre, fakts: Fakts):
    return ArkitektUnlok(
        rath=UnlokRath(
            link=UnlokLinkComposition(
                auth=HerreAuthLink(herre=herre),
                split=FaktsAIOHttpLink(fakts_group="lok", fakts=fakts),
            )
        )
    )



import os


def create_arkitekt_folder(with_cache: bool = True) -> str:
    """Creates the .arkitekt folder in the current directory.

    If the folder already exists, it does nothing.
    It automatically creates a .gitignore file, and a .dockerignore file,
    so that the Arkitekt credential files are not added to git.

    Parameters
    ----------
    with_cache : bool, optional
        Should we create a cache dir?, by default True

    Returns
    -------
    str
        The path to the .arkitekt folder.
    """
    os.makedirs(".arkitekt", exist_ok=True)
    if with_cache:
        os.makedirs(".arkitekt/cache", exist_ok=True)

    gitignore = os.path.join(".arkitekt", ".gitignore")
    dockerignore = os.path.join(".arkitekt", ".dockerignore")
    if not os.path.exists(gitignore):
        with open(gitignore, "w") as f:
            f.write(
                "# Hiding Arkitekt Credential files from git\n*.json\n*.temp\ncache/\nservers/"
            )
    if not os.path.exists(dockerignore):
        with open(dockerignore, "w") as f:
            f.write(
                "# Hiding Arkitekt Credential files from git\n*.json\n*.temp\ncache/\nservers/"
            )

    return os.path.abspath(".arkitekt")



from pydantic import BaseModel, Field, validator
import datetime
from typing import List, Optional, Union, Literal
from enum import Enum
import semver
import uuid
from rekuest.api.schema import DefinitionInput
from string import Formatter
import os

ALLOWED_BUILDER_KEYS = [
    "tag",
    "dockerfile",
]


class Framework(str, Enum):
    """Do we support other frameworks?"""

    VANILLA = "vanilla"
    TENSORFLOW = "tensorflow"
    PYTORCH = "pytorch"


class Requirement(str, Enum):
    """ """

    GPU = "gpu"


class Packager(str, Enum):
    CONDA = "conda"
    POETRY = "poetry"
    PIP = "pip"


class Manifest(BaseModel):
    identifier: str
    version: str
    author: str
    logo: Optional[str]
    entrypoint: str
    scopes: List[str]
    requirements: List[Requirement] = Field(default_factory=list)
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

    @validator("version", pre=True)
    def version_must_be_semver(cls, v) -> str:
        """Checks that the version is a valid semver version"""
        if isinstance(v, str):
            try:
                semver.VersionInfo.parse(v)
            except ValueError:
                raise ValueError("Version must be a valid semver version")
        return str(v)

    def to_console_string(self):
        return f"📦 {self.identifier} ({self.version}) by {self.author}"

    def to_builder_dict(self):
        return {
            "identifier": self.identifier,
            "version": self.version,
            "logo": self.logo,
            "scopes": self.scopes,
        }

    class Config:
        validate_assignment = True


class SelectorType(str, Enum):
    RAM = "ram"
    CPU = "cpu"
    GPU = "gpu"
    LABEL = "label"


class BaseSelector(BaseModel):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    required: bool = True

    class Config:
        extra = "forbid"

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector

        Should return a list of strings that can be used as docker params
        If the selector is not required, it should return an empty list

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return []

    def build_arkitekt_params(self) -> List[str]:
        """Builds the arkitekt params for this selector

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return []


class RAMSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["ram"]
    min: int


class CPUSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["cpu"]
    min: int
    frequency: Optional[int] = None


class CudaSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["cuda"]
    frequency: Optional[int] = Field(default=None, description="The frequency in MHz")
    memory: Optional[int] = Field(default=None, description="The memory in MB")
    architecture: Optional[str] = Field(
        default=None, description="The architecture of the GPU"
    )
    compute_capability: str = Field(
        default="3.5", description="The minimum compute capability"
    )
    cuda_cores: Optional[int] = None
    cuda_version: str = Field(default="10.2", description="The minimum cuda version")

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector

        Should return a list of strings that can be used as docker params
        If the selector is not required, it should return an empty list

        Returns
        -------
        List[str]
            The docker params for this selector
        """
        return [
            "--gpus",
            "all",
        ]


class RocmSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["rocm"]
    min: int
    frequency: Optional[int] = None
    memory: Optional[int] = None
    architecture: Optional[str] = None
    compute_capability: Optional[str] = None
    cuda_cores: Optional[int] = None
    cuda_version: Optional[str] = None

    def build_docker_params(self) -> List[str]:
        """Builds the docker params for this selector"""

        return [
            "--device=/dev/kfd",
            "--device=/dev/dri",
            "--group-add",
            "video",
            "--cap-add=SYS_PTRACE",
            "--security-opt",
            "seccomp=unconfined",
        ]


class LabelSelector(BaseSelector):
    """A selector is a way to describe a flavours preference for a
    compute node. It contains the node_id, the selector and the flavour_id.
    """

    type: Literal["label"]
    key: str
    value: str


class ServiceSelector(BaseSelector):
    """A service selector is a way to describe a flavours preference for a
    service. It contains the service_id,
    """

    type: Literal["service"]


Selector = Union[
    RAMSelector, CPUSelector, CudaSelector, RocmSelector, LabelSelector, ServiceSelector
]


class Inspection(BaseModel):
    definitions: List[DefinitionInput]
    size: int


class Flavour(BaseModel):
    selectors: List[Selector]
    description: str = Field(default="")
    dockerfile: str = Field(default="Dockerfile")
    build_command: List[str] = Field(
        default_factory=lambda: [
            "docker",
            "build",
            "-t",
            "{tag}",
            "-f",
            "{dockerfile}",
            ".",
        ]
    )

    @validator("build_command", each_item=True, always=True)
    def check_valid_template_name(cls, v):
        """Checks that the template name is valid"""
        for literal_text, field_name, format_spec, conversion in Formatter().parse(v):
            if field_name is not None:
                assert (
                    field_name in ALLOWED_BUILDER_KEYS
                ), f"Invalid template key {field_name}. Allowed keys are {ALLOWED_BUILDER_KEYS}"

        return v

    def generate_build_command(self, tag: str, relative_dir: str):
        """Generates the build command for this flavour"""

        dockerfile = os.path.join(relative_dir, self.dockerfile)

        return [v.format(tag=tag, dockerfile=dockerfile) for v in self.build_command]

    def check_relative_paths(self, flavour_folder: str):
        """Checks that the paths are relative to the flavour folder"""

        dockerfile_path = os.path.join(flavour_folder, self.dockerfile)

        if not os.path.exists(dockerfile_path):
            raise Exception(
                f"Could not find Dockerfile {self.dockerfile} in flavour {flavour_folder}"
            )


class Deployment(BaseModel):
    """A deployment is a Release of a Build.
    It contains the build_id, the manifest, the builder, the definitions, the image and the deployed_at timestamp.



    """

    deployment_run: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The unique identifier of the deployment run",
    )

    deployment_id: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The unique identifier of the deployment",
    )
    manifest: Manifest = Field(description="The manifest of the app that was deployed")
    selectors: List[Selector] = Field(
        description="The selectors are used to place this image on the nodes",
        default_factory=list,
    )
    flavour: str = Field(
        description="The flavour that was used to build this deployment",
        default="vanilla",
    )
    build_id: str = Field(
        description="The build_id of the build that was deployed. Is referenced in the build.yaml file."
    )
    inspection: Optional[Inspection] = Field(
        description="The inspection of the build that was deployed"
    )
    image: str = Field(
        description="The docker image that was built for this deployment"
    )
    deployed_at: datetime.datetime = Field(
        default_factory=datetime.datetime.now,
        description="The timestamp of the deployment",
    )


class DeploymentsConfigFile(BaseModel):
    """The ConfigFile is a pydantic model that represents the deployments.yaml file


    Parameters
    ----------
    BaseModel : _type_
        _description_
    """

    deployments: List[Deployment] = []
    latest_deployment_run: Optional[str] = None


class Build(BaseModel):
    build_run: str
    build_id: str
    inspection: Optional[Inspection] = None
    description: str = Field(default="")
    selectors: List[Selector] = Field(default_factory=list)
    flavour: str = Field(default="vanilla")
    manifest: Manifest
    build_at: datetime.datetime = Field(default_factory=datetime.datetime.now)
    base_docker_command: List[str] = Field(
        default_factory=lambda: ["docker", "run", "-it", "--net", "host"]
    )
    base_arkitekt_command: List[str] = Field(
        default_factory=lambda: ["arkitekt", "run", "prod", "--headless"]
    )

    def build_docker_command(self) -> List[str]:
        """Builds the docker command for this build"""

        base_command = self.base_docker_command

        for selector in self.selectors:
            base_command = base_command + selector.build_docker_params()

        base_command = base_command + [self.build_id]

        return base_command

    def build_arkitekt_command(self, fakts_url: str):
        """Builds the arkitekt command for this build"""

        base_command = self.base_arkitekt_command

        for selector in self.selectors:
            base_command = base_command + selector.build_arkitekt_params()

        base_command = base_command + ["--url", fakts_url]

        return base_command


class BuildsConfigFile(BaseModel):
    builds: List[Build] = Field(default_factory=list)
    latest_build_run: Optional[str] = None



from arkitekt import register
from mikro.api.schema import RepresentationFragment, from_xarray


@register
def max_intensity_projection(image: RepresentationFragment) -> RepresentationFragment:
    """Z-Project the Maximum Intensity

    This function projects the maximum intensity of the input image
    along the z-axis

    Parameters
    ----------
    image : RepresentationFragment
        The input image

    Returns
    -------
    RepresentationFragment
        The projected image

    """
    image = image.data.max(dim="z")
    return from_xarray(
        image, name="Max Intensity Projection" + image.name, origins=[image]
    )



import rich_click as click


import rich_click as click
from .remote import remote
from .local import local


@click.group()
@click.pass_context
def call(ctx):
    """Inspects your arkitekt app

    Inspects various parts of your arkitekt app. This is useful for debugging
    and development. It also represents methods that are called by the arkitekt
    server when you run your app in production mode.

    """


call.add_command(local, "local")
call.add_command(remote, "remote")



import rich_click as click
from .watch import watch
from .compile import compile
from .init import init


@click.group()
def gen():
    """Codegeneration tools for Arkitekt Apps (requires turms)

    Code generation for API's is done with the help of GraphQL Code Generation
    that is powered by [link=https://github.com/jhnnsrs/turms]turms[/link]. Simply
    design your API in the documents folder and run `arkitekt gen compile` to
    create fully typed code for your API. You can also run `arkitekt gen watch`
    to automatically generate code when your documents change. This is useful
    for development.

    """
    try:
        pass
    except ImportError as e:
        raise click.ClickException(
            "Turms is not installed. Please install turms first before using the arkitekt codegen."
        ) from e


gen.add_command(watch, "watch")
gen.add_command(compile, "compile")
gen.add_command(init, "init")



from importlib import import_module
from typing import Callable
from arkitekt.apps.types import App
import os


def import_builder(builder: str) -> Callable[..., App]:
    module_path, function_name = builder.rsplit(".", 1)
    module = import_module(module_path)
    function = getattr(module, function_name)
    return function


def build_relative_dir(*paths):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), *paths)



import rich_click as click
from arkitekt.cli.vars import get_manifest
import subprocess
from click import Context
from arkitekt.cli.io import get_builds


@click.command()
@click.option("--build", help="The build to use", type=str, default=None)
@click.option("--flavour", "-f", help="The flavour to use", default="vanilla")
@click.option(
    "--url", "-u", help="The fakts server to use", type=str, default="localhost:8000"
)
@click.option(
    "--builder", help="The builder to use", type=str, default="arkitekt.builders.easy"
)
@click.pass_context
def stage(ctx: Context, build: str, url: str, flavour: str, builder: str) -> None:
    """Stages the latest Build for testing

    Stages the current build for testing. This will create a temporary staged version
    of the app that is run agains the local arkitekt instance. The builder will be changed
    to the easy or provided builder to ensure that the app can be run headlessly


    """

    get_manifest(ctx)

    builds = get_builds(build)

    if len(builds) == 0:
        raise click.ClickException("Could not find any builds")

    if len(builds) > 1:
        try:
            build_model = next(
                build for build in builds.values() if build.flavour == flavour
            )
        except StopIteration:
            raise click.ClickException(
                f"Could not find a build for flavour {flavour}. Please run `arkitekt port build` "
                + "first to build the flavour"
            )

    else:
        build_model = list(builds.values())[0]

    command = build_model.build_docker_command() + build_model.build_arkitekt_command(
        url
    )

    click.echo(f"Staging build {build} with flavour {flavour} against {url}")
    click.echo("Running command: " + " ".join(command))

    subprocess.run(" ".join(command), shell=True)

    raise click.ClickException("Docker container exited")



import sys
import os

try:
    import rich_click as click

    from rich.console import Console
except ImportError:
    print(
        "Arkitekt CLI is not installed, please install it first. By installing the cli, e.g with `pip install arkitekt[cli]`, you can use the `arkitekt` command."
    )
    sys.exit(1)

from arkitekt.cli.vars import *
from arkitekt.cli.constants import *
from arkitekt.cli.texts import *
from arkitekt.cli.commands.run.main import run
from arkitekt.cli.commands.gen.main import gen
from arkitekt.cli.commands.server.main import server
from arkitekt.cli.commands.port.main import port
from arkitekt.cli.commands.init.main import init
from arkitekt.cli.commands.manifest.main import manifest
from arkitekt.cli.commands.inspect.main import inspect
from arkitekt.cli.commands.call.main import call
from arkitekt.cli.io import load_manifest
from arkitekt.utils import create_arkitekt_folder

default_docker_file = """
FROM python:3.8-slim-buster


RUN pip install arkitekt==0.4.23


RUN mkdir /app
COPY . /app
WORKDIR /app

"""


click.rich_click.HEADER_TEXT = LOGO
click.rich_click.ERRORS_EPILOGUE = ERROR_EPILOGUE
click.rich_click.USE_RICH_MARKUP = True


@click.group()
@click.pass_context
def cli(ctx):
    """Arkitekt is a framework for building safe and performant apps that then can be centrally orchestrated and managed
    in workflows.


    This is the CLI for the Arkitekt Python SDK. It allows you to create and deploy Arkitekt Apps from your python code
    as well as to run them locally for testing and development. For more information about Arkitekt, please visit
    [link=https://arkitekt.live]https://arkitekt.live[/link]
    """
    sys.path.append(os.getcwd())

    ctx.obj = {}
    console = Console()
    set_console(ctx, console)

    create_arkitekt_folder()

    manifest = load_manifest()
    if manifest:
        set_manifest(ctx, manifest)

    pass


cli.add_command(init, "init")
cli.add_command(run, "run")
cli.add_command(gen, "gen")
cli.add_command(port, "port")
cli.add_command(manifest, "manifest")
cli.add_command(inspect, "inspect")
cli.add_command(server, "server")
cli.add_command(call, "call")

if __name__ == "__main__":
    cli()



from arkitekt.cli.constants import compile_dockerfiles
from arkitekt.cli.types import Flavour
from arkitekt.cli.utils import build_relative_dir
import rich_click as click
from click import Context
from rich import get_console
from arkitekt.utils import create_arkitekt_folder
import yaml
from rich.panel import Panel

try:
    pass
except ImportError as e:
    raise ImportError("Please install rekuest to use this feature") from e

import os


@click.command()
@click.option("--flavour", "-f", help="The flavour to use", default="vanilla")
@click.option(
    "--description",
    "-d",
    help="The description for this flavour to use",
    default="This is a vanilla flavour",
)
@click.option(
    "--overwrite",
    "-o",
    help="Should we overwrite the existing Dockerfile?",
    is_flag=True,
    default=False,
)
@click.option(
    "--template",
    "-t",
    help="The dockerfile template to use",
    default="vanilla",
    type=click.Choice(compile_dockerfiles()),
)
@click.pass_context
def init(
    ctx: Context, description: str, overwrite: bool, flavour: str, template: str
) -> None:
    """Runs the port wizard to generate a dockerfile to be used with port"""

    arkitekt_folder = create_arkitekt_folder()

    flavour_folder = os.path.join(arkitekt_folder, "flavours", flavour)
    if os.path.exists(flavour_folder) and not overwrite:
        raise click.ClickException(
            f"The flavour {flavour} does already exist. Please initialize a different flavour or use the --overwrite flag"
        )
    else:
        os.makedirs(flavour_folder, exist_ok=True)

    config_file = os.path.join(flavour_folder, "config.yaml")
    dockerfile = os.path.join(flavour_folder, "Dockerfile")

    fl = Flavour(
        selectors=[],
        description=description,
        dockerfile="Dockerfile",
    )

    with open(config_file, "w") as file:
        yaml.dump(fl.dict(), file)

    with open(build_relative_dir("dockerfiles", f"{template}.dockerfile"), "r") as f:
        dockerfile_content = f.read()

    with open(dockerfile, "w") as f:
        f.write(dockerfile_content)

    panel = Panel(
        title=f"Created new flavour [bold]{flavour}[/bold]\n",
        renderable="You can now edit the Dockerfile and add selectors to the config.yaml file\n"
        + "To learn more about selectors and how flavours work, please visit [link=https://arkitekt.live]https://arkitekt.live[/link]",
        style="green",
    )

    get_console().print(panel)



import rich_click as click
from semver import (
    parse as parse_semver,
    bump_build,
    bump_major,
    bump_minor,
    bump_patch,
    bump_prerelease,
)
from arkitekt.cli.vars import get_console, get_manifest
from arkitekt.cli.io import write_manifest


@click.group()
@click.pass_context
def version(ctx):
    """Updates the version of the arkitekt app

    Arkitekt manifests versioning follow [link=https://semver.org]semver[/link] and are used to version the app.
    This provides an orthogonal way to version the app, beyond node versioning. The version is used to
    track changes and to provide a way to update the app in the platform. For more information, please visit
    [link=https://arkitekt.live]https://arkitekt.live[/link]

    """


@version.command("set")
@click.argument("VERSION", type=str, required=False)
@click.pass_context
def set_version(ctx, version):
    """Sets the version of the arkitekt app

    When setting the version, you can either provide a version, or you can let the cli
    prompt you for a version. If you provide a version, it will be parsed and validated
    against semver. If you let the cli prompt you, it will try to parse the current version
    and suggest a new version based on that. If the current version is not a valid semver
    version, it will prompt you for a new version without a suggestion.
    """

    manifest = get_manifest(ctx)
    get_console(ctx)
    old_version = manifest.version

    if not version:
        try:
            potential_new_version = bump_patch(old_version)
        except Exception:
            potential_new_version = None

        new_version = click.prompt(
            "Please provide a new version", default=potential_new_version, type=str
        )
        version = parse_semver(new_version)
        version = new_version

    manifest.version = version
    write_manifest(manifest)
    get_console().print(f"Version Updated from {old_version} to {version}")


@version.command()
@click.pass_context
def patch(ctx):
    """ "Patches the version of the arkitekt app


    Patches the version of the arkitekt app, by bumping the patch number.
    E.g. from 1.0.1 to 1.0.2. This should be used for bugfixes and small changes.
    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_patch(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def minor(ctx):
    """Bumps the minor version number of the arkitekt app

    Patches the version of the arkitekt app, by bumping the minor number.
    E.g. from 1.0.1 to 1.1.1. This should be used for new features, that
    are backwards compatible.

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_minor(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def major(ctx):
    """Increase the major version of the arkitekt app"

    Patches the version of the arkitekt app, by bumping the major number.
    E.g. from 1.0.1 to 2.0.1, This should be used for breaking changes,
    that are not backwards compatible (e.g. deleting a node).

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_major(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command()
@click.pass_context
def prerelease(ctx):
    """Patches the prerelease of the arkitekt app"


    Patches the version of the arkitekt app, by bumping the prerelease number.
    E.g. from 1.0.1 to 1.0.1-alpha.1
    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_prerelease(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")


@version.command("build")
@click.pass_context
def bump_build(ctx):
    """Patches the build of the arkitekt app

    Patches the version of the arkitekt app, by bumping the build number.
    E.g. from 1.0.1 to 1.0.1+1, This should be used for changes that are not
    reflected in the version number, but are still important to track (e.g. a
    hotfix).

    """
    manifest = get_manifest(ctx)
    console = get_console(ctx)
    old_version = manifest.version
    manifest.version = bump_build(old_version)
    write_manifest(manifest)
    console.print(f"Version Updated from {old_version} to {manifest.version}")



from arkitekt.healthz import FaktsHealthz
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.links.split import SplitLink
from rath.contrib.fakts.links.graphql_ws import FaktsGraphQLWSLink
from rath.contrib.herre.links.auth import HerreAuthLink
from omero_ark.rath import OmeroArkRathLinkComposition, OmeroArkRath
from omero_ark.omero_ark import OmeroArk
from graphql import OperationType

from fakts import Fakts
from herre import Herre


class ArkitektOmeroArk(OmeroArk):
    rath: OmeroArkRath
    healthz: FaktsHealthz


def build_arkitekt_omero_ark(
    fakts: Fakts, herre: Herre, fakts_group: str = "omero_ark"
) -> ArkitektOmeroArk:
    rath = OmeroArkRath(
        link=OmeroArkRathLinkComposition(
            auth=HerreAuthLink(herre=herre),
            split=SplitLink(
                left=FaktsAIOHttpLink(fakts_group=fakts_group, fakts=fakts),
                right=FaktsGraphQLWSLink(fakts_group=fakts_group, fakts=fakts),
                split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
            ),
        )
    )

    return ArkitektOmeroArk(
        rath=rath,
        healthz=FaktsHealthz(fakts_group=fakts_group, fakts=fakts),
    )



from ssl import SSLContext
import ssl
from typing import Optional

import aiohttp
import certifi
from pydantic import Field
from koil.composition import KoiledModel
from fakts.fakts import Fakts


class FaktsHealthz(KoiledModel):
    """A model to check the healthz endpoint on a service configured through fakts."""

    fakts: Fakts = Field(
        description="The Fakts instance to use to fetch the healthz endpoint"
    )
    fakts_group: str = Field(
        description="The Fakts group to use to fetch the healthz endpoint"
    )
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        description="SSL Context to use for the request",
    )

    healthz_item: str = Field(
        default="healthz",
        description="The item in the Fakts group that contains the healthz endpoint. Fakts->Group->Item",
    )
    strict: bool = Field(
        default=False,
        description="Whether to raise an exception if the healthz endpoint is not working",
    )
    endpoint_url: Optional[str] = Field(description="Overwrite the endpoint url.")

    async def check(self):
        """Check the healthz endpoint of the Fakts instance."""
        self.endpoint_url = await self.fakts.aget(self.fakts_group)[self.healthz_item]

        async with aiohttp.ClientSession(
            headers={"Content-Type": "application/json"},
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            # get json from endpoint
            async with session.get(self.endpoint_url + "?format=json") as resp:
                healthz_json = await resp.json()
                if self.strict:
                    # check if all services are working
                    faulty_services = []
                    for key, value in healthz_json.items():
                        if value != "working":
                            faulty_services.append({"service": key, "error": value})

                    if faulty_services:
                        raise Exception(f"Faulty services: {faulty_services}")

                if "detail" in healthz_json:
                    # detail is only present if the endpoint is not working
                    raise Exception("Error on the Healtzh endpoint ")

                return healthz_json

    class Config:
        underscore_attrs_are_private = True



import rich_click as click
from .init import init
from .up import up
from .down import down
from click import Context


@click.group()
@click.pass_context
def server(ctx: Context) -> None:
    """Manages your arkitekt deployments

    You can directly deploy your app to a server with the arkitekt server command. This will
    create a local docker-compose.yml file and a .arkitekt folder in your project. You can then
    use the arkitekt server build command to build your app for port and the arkitekt server run
    command to run your app on the server.

    """

    pass


server.add_command(init, "init")
server.add_command(up, "up")
server.add_command(down, "down")



""" A deployed kluster instance package"""

from dokker import Deployment
import os
from koil.composition import Composition
from arkitekt.apps.types import EasyApp
from arkitekt.builders import easy
from dokker.projects.contrib.konstruktor import KonstruktorProject
from typing import Optional
from arkitekt.constants import KONSTRUKTOR_URL


class ArkitektDeployment(Deployment):
    project: KonstruktorProject


def build_deployment(
    channel: str, name: Optional[str] = None, repo_url: str = KONSTRUKTOR_URL
) -> ArkitektDeployment:
    """Builds a deploymen of kluster for testing

    This will return a deployment of kluster that is ready to be used for testing.
    It will have a health check that will check the graphql endpoint.

    Returns
    -------
    Deployment
        The deployment of kluster (dokker.Deployment)


    """
    project = KonstruktorProject(
        channel=channel,
        repo_url=repo_url,
        name=name,
        base_dir=os.path.join(os.getcwd(), ".arkitekt", "deployments"),
    )

    deployment = ArkitektDeployment(project=project)
    deployment.add_health_check(
        url="http://localhost:8000/ht",
        service="lok",
        timeout=5,
        max_retries=10,
    )
    deployment.health_on_enter = True
    return deployment


async def token_loader() -> str:
    """Returns a token as defined in the static_token setting for kluster"""
    return "demo"


class DeployedArkitekt(Composition):
    """A deployed kluster instance

    THis is a composition of both a deployment of kluster-server
    (and kluster-gateway) and a client for that deployment. It is
    the fastest way to get a fully functioning kluster instance,
    ready for testing.


    """

    deployment: ArkitektDeployment
    app: EasyApp


def deployed(channel: str, *args, **kwargs) -> DeployedArkitekt:
    """Create a deployed kluster

    A deployed kluster is a composition of a deployment of the
    kluster server and a kluster client.
    This means a fully functioning kluster instance will be spun up when
    the context manager is entered.

    To inspect the deployment, use the `deployment` attribute.
    To interact with the kluster, use the `kluster` attribute.


    Returns
    -------
    DeployedKluster
        The deployed kluster instance (Composition)
    """

    url = "localhost:8000"
    # TODO: Retrieve the url from the deployment

    return DeployedArkitekt(
        deployment=build_deployment(channel),
        app=easy(*args, url=url, **kwargs),
    )



"""
This module contains the types for the apps
depending on the builder used.

This module imports all the apps and their types
and sets them as attributes on the App class, if they are available.
If they are not available, they are set to Any, so that we can add 
an import exception to the app.


"""
from koil.composition import Composition
from typing import TYPE_CHECKING
from arkitekt.model import Manifest
from arkitekt.apps.fallbacks import ImportException

import logging


logger = logging.getLogger(__name__)

try:
    from arkitekt.apps.service.rekuest import ArkitektRekuest
except ImportError:
    ArkitektRekuest = ImportException  # type: ignore


try:
    from arkitekt.apps.service.kluster import ArkitektKluster
except ImportError:
    ArkitektKluster = ImportException  # type: ignore


try:
    from arkitekt.apps.service.rekuest_next import ArkitektRekuestNext
except ImportError:
    ArkitektRekuestNext = ImportException  # type: ignore

try:
    from arkitekt.apps.service.fakts_qt import ArkitektFaktsQt
except ImportError:
    ArkitektFaktsQt = ImportException

try:
    from arkitekt.apps.service.herre_qt import ArkitektHerreQt
except ImportError:
    ArkitektHerreQt = ImportException


try:
    from arkitekt.apps.service.mikro import ArkitektMikro
except ImportError:
    ArkitektMikro = ImportException  # type: ignore
try:
    from arkitekt.apps.service.mikro_next import ArkitektMikroNext
except ImportError:
    ArkitektMikroNext = ImportException  # type: ignore
try:
    from arkitekt.apps.service.herre import ArkitektHerre
except ImportError:
    ArkitektHerre = ImportException  # type: ignore
try:
    from arkitekt.apps.service.fluss import ArkitektFluss
except ImportError:
    ArkitektFluss = ImportException  # type: ignore
try:
    from arkitekt.apps.service.unlok import ArkitektUnlok
except ImportError:
    ArkitektUnlok = ImportException  # type: ignore
try:
    from arkitekt.apps.service.omero_ark import ArkitektOmeroArk
except ImportError:
    ArkitektOmeroArk = ImportException  # type: ignore


try:
    from arkitekt.apps.service.fakts import ArkitektFakts
except ImportError:
    ArkitektFakts = ImportException  # type: ignore


try:
    from arkitekt.apps.service.fakts_next import ArkitektFaktsNext, Manifest
except ImportError:
    ArkitektFaktsNext = ImportException  # type: ignore


if TYPE_CHECKING:
    from arkitekt.apps.service.fakts import ArkitektFakts
    from arkitekt.apps.service.fakts_next import ArkitektFaktsNext
    from arkitekt.apps.service.rekuest import ArkitektRekuest
    from arkitekt.apps.service.omero_ark import ArkitektOmeroArk
    from arkitekt.apps.service.herre import ArkitektHerre
    from arkitekt.apps.service.mikro import ArkitektMikro
    from arkitekt.apps.service.mikro_next import ArkitektMikroNext
    from arkitekt.apps.service.fluss import ArkitektFluss
    from arkitekt.apps.service.unlok import ArkitektUnlok
    from arkitekt.apps.service.rekuest_next import ArkitektRekuestNext
    from arkitekt.apps.service.herre_qt import ArkitektHerreQt
    from arkitekt.apps.service.kluster import ArkitektKluster
    from arkitekt.apps.service.fakts_qt import ArkitektFaktsQt


class App(Composition):
    """An app that is built with the easy builder"""

    manifest: Manifest
    fakts: ArkitektFakts
    herre: ArkitektHerre
    rekuest: ArkitektRekuest
    mikro: ArkitektMikro
    fluss: ArkitektFluss
    unlok: ArkitektUnlok
    omero_ark: ArkitektOmeroArk
    kluster: ArkitektKluster


class EasyApp(Composition):
    """An app that is built with the easy builder"""

    manifest: Manifest
    fakts: ArkitektFakts
    herre: ArkitektHerre
    rekuest: ArkitektRekuest
    mikro: ArkitektMikro
    fluss: ArkitektFluss
    unlok: ArkitektUnlok
    omero_ark: ArkitektOmeroArk
    kluster: ArkitektKluster


class NextApp(Composition):
    """An app that is built with the next builder"""

    manifest: Manifest
    fakts: ArkitektFaktsNext
    herre: ArkitektHerre
    rekuest: ArkitektRekuestNext
    mikro: ArkitektMikroNext
    fluss: ArkitektFluss
    unlok: ArkitektUnlok
    omero_ark: ArkitektOmeroArk
    kluster: ArkitektKluster


class QtApp(Composition):
    manifest: Manifest
    fakts: ArkitektFaktsQt
    herre: ArkitektHerreQt
    rekuest: ArkitektRekuest
    mikro: ArkitektMikro
    unlok: ArkitektUnlok
    fluss: ArkitektFluss
    kluster: ArkitektKluster



def missing_rekuest_install(*args, **kwargs):
    raise ImportError("Missing import: rekuest. Please install the missing package. ")


def missing_rekuest_next_install(*args, **kwargs):
    raise ImportError(
        "Missing import: rekuest_next. Please install the missing package. "
    )


try:
    from rekuest.register import register, register_structure, PortGroupInput as group
    from rekuest.actors.reactive.api import (
        log,
        alog,
        progress,
        aprogress,
        useGuardian,
        useInstanceID,
        useUser,
    )

except ImportError:
    register = missing_rekuest_install
    register_structure = missing_rekuest_install
    group = missing_rekuest_install
    log = missing_rekuest_install
    alog = missing_rekuest_install
    progress = missing_rekuest_install
    aprogress = missing_rekuest_install
    useGuardian = missing_rekuest_install
    useInstanceID = missing_rekuest_install
    useUser = missing_rekuest_install


try:
    from rekuest_next.register import register as register_next
except ImportError:
    register_next = missing_rekuest_next_install

from .builders import easy, publicqt, jupy, scheduler, next
from .apps.types import App

__all__ = [
    "App",
    "register",
    "easy",
    "publicqt",
    "jupy",
    "log",
    "alog",
    "progress",
    "aprogress",
    "scheduler",
    "register_structure",
    "group",
    "useGuardian",
    "useInstanceID",
    "useUser",
    "next",
]



import rich_click as click
import os
from arkitekt.cli.constants import *
from getpass import getuser
from arkitekt.cli.types import Manifest, Requirement
from arkitekt.cli.vars import get_manifest, get_console
import semver
from arkitekt.cli.io import write_manifest
from rich.panel import Panel
from typing import Optional, List


def check_overwrite_app(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    app_file = ctx.params["entrypoint"] + ".py"
    if os.path.exists(app_file) and not value:
        should_overwrite = click.confirm(
            "Entrypoint File already exists. Do you want to overwrite?"
        )
        return should_overwrite

    return value


def check_overwrite(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    try:
        manifest = get_manifest(ctx)
        if not value:
            should_overwrite = click.confirm(
                f"Another Arkitekt app {manifest.to_console_string()} exists already at {os.getcwd()}?. Do you want to overwrite?",
                abort=True,
            )
            if not should_overwrite:
                ctx.abort()
    except click.ClickException:
        pass

    return value


def ensure_semver(ctx, param, value):
    """Callback to check and prompt for file overwrite."""

    if not value:
        value = click.prompt(
            "The version of your app",
            default="0.0.1",
        )

    while not semver.Version.is_valid(value):
        get_console().print(
            "Arkitekt versions need to follow [link=https://semver.org]semver[/link]. Please choose a correct format (examples: 0.0.0, 0.1.0, 0.0.0-alpha.1)"
        )
        value = click.prompt(
            "The version of your app",
            default="0.0.1",
        )

    return value


@click.command()
@click.option(
    "--overwrite-manifest",
    "-om",
    help="Should we overwrite the existing manifest if it already exists?",
    is_flag=True,
    default=False,
    callback=check_overwrite,
)
@click.option(
    "--template",
    "-t",
    help="The template to use. You can choose from a variety of preconfigured templates. They are just starting points and can be changed later.",
    type=click.Choice(compile_templates()),
    default="simple",
)
@click.option(
    "--identifier",
    "-i",
    help="The identifier of your app. This will be used to identify your app in the Arkitekt ecosystem. It should be unique and should follow the [link=https://en.wikipedia.org/wiki/Reverse_domain_name_notation]reverse domain name notation[/link] (example: com.example.myapp)",
    prompt="Your app identifier",
    default=os.path.basename(os.getcwd()),
)
@click.option(
    "--version",
    "-v",
    help="The version of your app. Needs to follow [link=https://semver.org/]semantic versioning[/link].",
    default="0.0.1",
    callback=ensure_semver,
)
@click.option(
    "--author",
    help="The author of your app. This will be shown to users of your app",
    prompt="Your name",
    default=getuser(),
)
@click.option(
    "--template",
    help="Which template to use top create entrypoint",
    prompt="Your app file template",
    type=click.Choice(compile_templates()),
    default="simple",
)
@click.option(
    "--logo",
    help="Which logo to use for this app, needs to be a valid url",
    required=False,
)
@click.option(
    "--entrypoint",
    "-e",
    help="The entrypoint of your app. This will be the name of the python file. Omit the .py ending",
    prompt="Your app file",
    default="app",
)
@click.option(
    "--overwrite-app",
    "-oa",
    help="Do you want to overwrite the app file if it exists?",
    is_flag=True,
    default=False,
    callback=check_overwrite_app,
)
@click.option(
    "--requirements",
    "-r",
    help="Dedicated hardware requirements for this app. They will be used to inform Plugin Schedulers like port, on which compute resources to put your plugin app, once deployed.You can choose multiple for your app. For a list of requirements, run `arkitekt manifest requirements available`",
    type=click.Choice(compile_requirements()),
    multiple=True,
    default=[],
)
@click.option(
    "--scopes",
    "-s",
    help="The scopes of the app. You can choose multiple for your app. For a list of scopes, run `arkitekt manifest scopes available`",
    type=click.Choice(compile_scopes()),
    multiple=True,
    default=["read"],
)
@click.pass_context
def init(
    ctx,
    identifier: str,
    version: str,
    author: str,
    logo: Optional[str],
    scopes: List[str],
    template: str,
    requirements: List[Requirement],
    entrypoint: str,
    overwrite_manifest: bool,
    overwrite_app: bool,
):
    """Initializes an Arkitekt app

    This command will create a new Arkitekt app in the current directory. It will
    create a `.arkitekt` folder that will contain a manifest and a `app.py` file,
    which will serve as the entrypoint for your app. By default, the app will be
    initialized with a simple hello world app, but you can choose from a variety
    of templates.

    """

    console = get_console(ctx)

    manifest = Manifest(
        logo=logo,
        author=author,
        identifier=identifier,
        version=version,
        scopes=scopes,
        requirements=requirements,
        entrypoint=entrypoint,
    )

    with open(build_relative_dir("templates", f"{template}.py")) as f:
        template_app = f.read()

    if not os.path.exists(f"{entrypoint}.py") or overwrite_app:
        with open(f"{entrypoint}.py", "w") as f:
            f.write(template_app)

    write_manifest(manifest)
    md = Panel(
        f"{manifest.to_console_string()} was successfully initialized\n\n"
        + "[not bold white]We are excited to see what you come up with!",
        border_style="green",
        style="green",
    )
    console.print(md)



import rich_click as click
from importlib import import_module
from arkitekt.cli.vars import get_console, get_manifest
from arkitekt.cli.options import with_builder
import json


async def run_app(app):
    async with app:
        await app.rekuest.run()


@click.command("prod")
@click.pass_context
@click.option(
    "--pretty",
    "-p",
    help="Should we just output json?",
    is_flag=True,
    default=False,
)
@with_builder
def definitions(
    ctx,
    pretty: bool,
    builder=None,
):
    """Runs the app in production mode

    \n
    You can specify the builder to use with the --builder flag. By default, the easy builder is used, which is designed to be easy to use and to get started with.

    """
    from rekuest.definition.registry import get_default_definition_registry

    manifest = get_manifest(ctx)
    console = get_console(ctx)

    entrypoint = manifest.entrypoint

    with console.status("Loading entrypoint module..."):
        try:
            import_module(entrypoint)
        except ModuleNotFoundError as e:
            console.print(f"Could not find entrypoint module {entrypoint}")
            raise e

    definitions = get_default_definition_registry().dump()

    if pretty:
        console.print(json.dumps(definitions, indent=2))
    else:
        print(json.dumps(definitions))



from fakts.fakts import Fakts
from typing import Optional
from fakts.grants.remote import RemoteGrant
from fakts.grants.remote.demanders.auto_save import AutoSaveDemander
from fakts.grants.remote.discovery.auto_save import AutoSaveDiscovery
from fakts.grants.remote.discovery.qt.auto_save_endpoint_widget import (
    AutoSaveEndpointWidget,
)
from fakts.grants.remote.discovery.qt.qt_settings_endpoint_store import (
    QtSettingsEndpointStore,
)
from fakts.grants.remote.demanders.qt.qt_settings_token_store import QTSettingTokenStore

from fakts.grants.remote.demanders.retrieve import RetrieveDemander
from fakts.grants.remote.claimers.post import ClaimEndpointClaimer
from fakts.grants.remote.discovery.qt.selectable_beacon import (
    SelectBeaconWidget,
    QtSelectableDiscovery,
)
from arkitekt.model import Manifest
from qtpy import QtCore, QtWidgets


class ArkitektFaktsAutoSaveDiscovery(AutoSaveDiscovery):
    """An Arkitekt Fakts discovery that uses Qt widgets for token and endpoint storage"""

    discovery: QtSelectableDiscovery


class ArkitektFaktsQtRemoteGrant(RemoteGrant):
    """An Arkitekt Fakts grant that uses Qt widgets for token and endpoint storage"""

    discovery: ArkitektFaktsAutoSaveDiscovery


class ArkitektFaktsQt(Fakts):
    """A Fakts that uses Qt widgets for token and endpoint storage"""

    grant: ArkitektFaktsQtRemoteGrant


def build_arkitekt_qt_fakts(
    manifest: Manifest,
    no_cache: Optional[bool] = False,
    beacon_widget: Optional[QtWidgets.QWidget] = None,
    parent: Optional[QtWidgets.QWidget] = None,
    settings: Optional[QtCore.QSettings] = None,
) -> ArkitektFaktsQt:
    beacon_widget = beacon_widget or SelectBeaconWidget(
        parent=parent, settings=settings
    )

    return ArkitektFaktsQt(
        grant=RemoteGrant(
            demander=AutoSaveDemander(
                store=QTSettingTokenStore(
                    settings=settings,
                    save_key="fakts_token",
                ),
                demander=RetrieveDemander(
                    manifest=manifest,
                    redirect_uri="http://localhost:6767",
                ),
            ),
            discovery=AutoSaveDiscovery(
                store=QtSettingsEndpointStore(
                    settings=settings,
                    save_key="fakts_endpoint",
                ),
                decider=AutoSaveEndpointWidget(
                    parent=parent,
                ),
                discovery=QtSelectableDiscovery(
                    widget=beacon_widget,
                    settings=settings,
                    allow_appending_slash=True,
                    auto_protocols=["http", "https"],
                ),
            ),
            claimer=ClaimEndpointClaimer(),
        )
    )



import rich_click as click
from click import Context
from dokker.projects.local import LocalProject
from dokker.loggers.print import PrintLogger
from dokker.deployment import Deployment
from typing import Optional
import os
from .utils import compile_options


@click.command()
@click.option(
    "--name",
    help="The name of the deployment",
    default=None,
    required=False,
    type=click.Choice(compile_options()),
)
@click.pass_context
def down(
    ctx: Context,
    name: Optional[str] = None,
) -> None:
    """
    Initializes a new deployment from a konstruktor template.
    this will create a new folder in the .arkitekt/deployments folder
    with the name of the deployment and the docker-compose.yml file

    """
    if not name:
        options = compile_options()
        if not options:
            raise click.ClickException(
                "No deployments found. Please run arkitekt server init first"
            )

        name = options[0]

    print(f"Running {name}")

    project = LocalProject(
        compose_files=[
            os.path.join(
                os.getcwd(), ".arkitekt", "deployments", name, "docker-compose.yaml"
            )
        ],
    )

    deployment = Deployment(project=project, logger=PrintLogger())
    deployment.up_on_enter = False
    deployment.down_on_exit = True

    with deployment:
        print("Shutting down...")

    print("Done")



"""
This modules provides the main app. It is responsible for setting up the connection to the mikro-server and
handling authentification and setting up the configuration. Mikro handles the creation of the datalayer and
the graphql client.

You can compose this app together with other apps to create a full fledged app. (Like combining with
arkitekt to enable to call functions that you define on the app)

Example:

    A simple app that takes it configuraiton from basic.fakts and connects to the mikro-server.
    You can define all of the logic within the context manager

    ```python
    from mikro.app import MikroApp

    app = MikroApp(fakts=Fakts(subapp="basic"))

    with app:
        # do stuff

    ```

    Async Usage:


    ```python
    from mikro.app import MikroApp

    app = MikroApp(fakts=Fakts(subapp="basic"))

    async with app:
        # do stuff

    ```

 
"""


from arkitekt.healthz import FaktsHealthz
from kluster.kluster import Kluster
from kluster.rath import KlusterRath, KlusterRathLinkComposition
from rath.links.split import SplitLink
from rath.contrib.fakts.links.aiohttp import FaktsAIOHttpLink
from rath.contrib.fakts.links.graphql_ws import FaktsGraphQLWSLink
from rath.contrib.herre.links.auth import HerreAuthLink
from graphql import OperationType
from herre import Herre
from fakts import Fakts
from kluster.contrib.arkitekt_repository import ArkitektRepository


class ArkitektKluster(Kluster):
    rath: KlusterRath
    repo: ArkitektRepository
    healthz: FaktsHealthz


def build_arkitekt_kluster(herre: Herre, fakts: Fakts):
    repo = ArkitektRepository(
        fakts=fakts,
        herre=herre,
        fakts_key="kluster.gateway_url",
    )

    return ArkitektKluster(
        rath=KlusterRath(
            link=KlusterRathLinkComposition(
                auth=HerreAuthLink(herre=herre),
                split=SplitLink(
                    left=FaktsAIOHttpLink(fakts_group="kluster", fakts=fakts),
                    right=FaktsGraphQLWSLink(fakts_group="kluster", fakts=fakts),
                    split=lambda o: o.node.operation != OperationType.SUBSCRIPTION,
                ),
            )
        ),
        healthz=FaktsHealthz(fakts_group="kluster", fakts=fakts),
        repo=repo,
    )



import os
import shutil
import rich_click as click
from arkitekt.cli.options import (
    with_documents,
    with_graphql_config,
    with_api_path,
    with_boring,
    with_choose_services,
    with_schemas,
    with_seperate_document_dirs,
)
import yaml
from arkitekt.cli.utils import build_relative_dir
from arkitekt.cli.vars import get_console, get_manifest


@click.command()
@with_seperate_document_dirs
@with_boring
@with_choose_services
@with_graphql_config
@with_api_path
@with_schemas
@with_graphql_config
@with_documents
@click.pass_context
def init(ctx, boring, services, config, documents, schemas, path, seperate_doc_dirs):
    """Initialize code generation for the arkitekt app

    Code generation for API's is done with the help of GraphQL Code Generation
    that is powered by turms. This command initializes the code generation for
    the app. It creates the necessary folders and files for the code generation
    to work. It also creates a graphql config file that is used by turms to
    generate the code.

    """
    app_directory = os.getcwd()

    app_api_path = os.path.join(app_directory, path)
    app_documents = os.path.join(app_directory, "graphql", "documents")

    app_schemas = os.path.join(app_directory, "graphql", "schemas")

    if documents:
        os.makedirs(app_documents, exist_ok=True)
    if schemas:
        os.makedirs(app_schemas, exist_ok=True)
    if path:
        os.makedirs(app_api_path, exist_ok=True)

    # Initializing the config
    projects = {}

    base_config = yaml.load(
        open(build_relative_dir("configs", "base.yaml"), "r"), Loader=yaml.FullLoader
    )

    for service in services:
        schema_path = build_relative_dir("schemas", f"{service}.schema.graphql")

        if documents:
            os.makedirs(os.path.join(app_documents, service), exist_ok=True)
            if seperate_doc_dirs:
                os.makedirs(
                    os.path.join(app_documents, service, "queries"), exist_ok=True
                )
                os.makedirs(
                    os.path.join(app_documents, service, "mutations"), exist_ok=True
                )
                os.makedirs(
                    os.path.join(app_documents, service, "subscriptions"), exist_ok=True
                )

        if schemas:
            if os.path.exists(schema_path):
                try:
                    shutil.copyfile(
                        schema_path,
                        os.path.join(app_schemas, service + ".graphql"),
                    )
                except FileExistsError:
                    if click.confirm(
                        f"Schema for {service} already exist. Do you want to overwrite them?"
                    ):
                        shutil.copyfile(
                            schema_path,
                            os.path.join(app_schemas, service + ".graphql"),
                        )
            else:
                get_console(ctx).print(f"[red]No schema found for {service} [/]")

        try:
            project = base_config["projects"][service]
        except KeyError:
            get_console(ctx).print(f"[red]No config found for {service} [/]")
            continue

        if schemas:
            project["schema"] = os.path.join(app_schemas, service + ".graphql")
        if documents:
            project["documents"] = (
                os.path.join(app_documents, service) + "/**/*.graphql"
            )

        project["extensions"]["turms"]["out_dir"] = path
        project["extensions"]["turms"]["generated_name"] = f"{service}.py"

        projects[service] = project

    if os.path.exists(config):
        if not click.confirm(
            f"GraphQL Config file already exists. Do you want to overwrite?"
        ):
            click.echo("Aborting...")
            ctx.abort()

    graph_config_path = os.path.join(app_directory, config)
    yaml.safe_dump(
        {"projects": projects}, open(graph_config_path, "w"), sort_keys=False
    )
    get_console(ctx).print(f"Config file written to {graph_config_path}")



from .types import Requirement
from .utils import build_relative_dir
from typing import List
import os


def compile_scopes() -> List[str]:
    """Compile all available scopes"""
    return ["read", "write"]


def compile_requirements():
    """Compile all available requirements"""
    return [Requirement.GPU.value]


def compile_builders():
    return ["arkitekt.builders.easy", "arkitekt.builders.port"]


def compile_runtimes():
    return ["nvidia", "standard"]


def compile_schema_versions() -> List[str]:
    z = build_relative_dir("schemas")
    return [
        os.path.basename(f) for f in os.listdir(z) if os.path.isdir(os.path.join(z, f))
    ]


def compile_configs() -> List[str]:
    z = build_relative_dir("configs")
    return [
        os.path.basename(f) for f in os.listdir(z) if os.path.isdir(os.path.join(z, f))
    ]


def compile_dockerfiles() -> List[str]:
    z = build_relative_dir("dockerfiles")
    return [
        os.path.basename(f).replace(".dockerfile", "")
        for f in os.listdir(z)
        if os.path.isfile(os.path.join(z, f))
    ]


def compile_templates() -> List[str]:
    z = build_relative_dir("templates")
    return [
        os.path.basename(f).split(".")[0]
        for f in os.listdir(z)
        if os.path.isfile(os.path.join(z, f))
    ]


def compile_services() -> List[str]:
    z = build_relative_dir("schemas")
    return [
        os.path.basename(f).split(".")[0]
        for f in os.listdir(z)
        if os.path.isfile(os.path.join(z, f))
    ]



import rich_click as click


import rich_click as click
from .variables import variables
from .definitions import definitions


@click.group()
@click.pass_context
def inspect(ctx):
    """Inspects your arkitekt app

    Inspects various parts of your arkitekt app. This is useful for debugging
    and development. It also represents methods that are called by the arkitekt
    server when you run your app in production mode.

    """


inspect.add_command(variables, "variables")
inspect.add_command(definitions, "definitions")



from arkitekt.model import Manifest
from .types import NextApp
from arkitekt.apps.fallbacks import ImportException
from arkitekt.apps.service.fakts_next import build_arkitekt_fakts_next
from arkitekt.apps.service.herre import build_arkitekt_herre


def build_next_app(
    manifest: Manifest,
    url=None,
    no_cache=False,
    headless=False,
    instance_id=None,
    token=None,
    app_kind="development",
):
    fakts = build_arkitekt_fakts_next(
        manifest=manifest,
        url=url,
        no_cache=no_cache,
        headless=headless,
        client_kind=app_kind,
    )

    herre = build_arkitekt_herre(fakts=fakts)

    try:
        from arkitekt.apps.service.rekuest_next import build_arkitekt_rekuest_next

        rekuest = build_arkitekt_rekuest_next(
            fakts=fakts, herre=herre, instance_id=instance_id
        )
    except ImportError as e:
        rekuest = ImportException(import_exception=e, install_library="rekuest_next")

    try:
        from arkitekt.apps.mikro_next import build_arkitekt_mikro_next

        mikro = build_arkitekt_mikro_next(fakts=fakts, herre=herre)
    except ImportError as e:
        mikro = ImportException(import_exception=e, install_library="mikro_next")

    try:
        from arkitekt.apps.service.unlok import build_arkitekt_unlok

        unlok = build_arkitekt_unlok(herre=herre, fakts=fakts)
    except ImportError as e:
        unlok = ImportException(import_exception=e, install_library="unlok")

    try:
        from arkitekt.apps.service.fluss import build_arkitekt_fluss

        fluss = build_arkitekt_fluss(herre=herre, fakts=fakts)
    except ImportError as e:
        fluss = ImportException(import_exception=e, install_library="fluss")

    try:
        from arkitekt.apps.service.omero_ark import build_arkitekt_omero_ark

        omero_ark = build_arkitekt_omero_ark(herre=herre, fakts=fakts)
    except ImportError as e:
        omero_ark = ImportException(import_exception=e, install_library="omero_ark")

    try:
        from arkitekt.apps.service.kluster import build_arkitekt_kluster

        kluster = build_arkitekt_kluster(herre=herre, fakts=fakts)
    except ImportError as e:
        kluster = ImportException(import_exception=e, install_library="kluster")

    return NextApp(
        manifest=manifest,
        fakts=fakts,
        herre=herre,
        rekuest=rekuest,
        mikro=mikro,
        unlok=unlok,
        fluss=fluss,
        kluster=kluster,
        omero_ark=omero_ark,
    )



import rich_click as click
from rich.table import Table
from rich.panel import Panel
from rich.console import Group
from arkitekt.cli.vars import get_manifest, get_console


@click.command()
@click.pass_context
def inspect(ctx) -> None:
    """Inspect the [i]current[/i] manifest of this app

    The manifest is used to describe the app and its rights (scopes) and requirements, to be run on the platform.
    This manifest is used to authenticate the app with the platform establishing its scopes and requirements.


    """
    manifest = get_manifest(ctx)

    table = Table.grid()
    table.add_column()
    table.add_column()
    table.add_row("Identifier", manifest.identifier)
    table.add_row("Version", manifest.version)
    table.add_row("Author", manifest.author)
    table.add_row("Logo", manifest.logo or "-")
    table.add_row("Entrypoint", manifest.entrypoint)
    table.add_row("Scopes", ", ".join(manifest.scopes) if manifest.scopes else "-")
    table.add_row(
        "Requirements",
        ", ".join(manifest.requirements) if manifest.requirements else "-",
    )
    table.add_row("Created at", str(manifest.created_at.strftime("%Y/%m/%d %H:%M")))

    panel = Panel(
        Group("[bold green]Manifest[/]", table),
        title_align="center",
        border_style="green",
        style="white",
    )

    get_console(ctx).print(panel)



import rich_click as click
from click import Context
from dokker.projects.contrib.konstruktor import KonstruktorProject
import asyncio
from typing import Optional
import os

DEFAULT_REPO_URL = (
    "https://raw.githubusercontent.com/jhnnsrs/konstruktor/master/repo/channels.json"
)


@click.command()
@click.option("--channel", help="The channel to use", default="paper")
@click.option("--name", help="The name of the deployment", default=None, required=False)
@click.option("--overwrite", help="The repo to use", default=False, is_flag=True)
@click.option("--repo-url", help="The repo to use", default=DEFAULT_REPO_URL)
@click.option(
    "--setup",
    "-s",
    "setup",
    help="Key Value pairs for the setup",
    type=(str, str),
    multiple=True,
)
@click.pass_context
def init(
    ctx: Context,
    channel: str,
    repo_url: str,
    overwrite: bool = False,
    setup: Optional[list] = None,
    name: Optional[str] = None,
) -> None:
    """
    Initializes a new deployment from a konstruktor template.
    this will create a new folder in the .arkitekt/deployments folder
    with the name of the deployment and the docker-compose.yml file

    """

    extra_content = dict(setup or [])

    project = KonstruktorProject(
        channel=channel,
        repo_url=repo_url,
        overwrite_if_exists=overwrite,
        name=name,
        extra_context=extra_content,
        base_dir=os.path.join(os.getcwd(), ".arkitekt", "deployments"),
    )

    asyncio.run(project.ainititialize())
